import os


class BaseFilter(object):
    def __init__(self, filter_string):
        self.filter_string = filter_string

    def apply(self, tree, options):
        raise NotImplementedError(tree, options)


class IgnoreDir(BaseFilter):
    def __init__(self, ignore_dirs):
        if not isinstance(ignore_dirs, (list, tuple)):
            ignore_dirs = [ignore_dirs]
        self.ignore_dirs = ignore_dirs
        super().__init__('UNUSED')

    def apply(self, tree, options):
        rv = {}
        for dirname in tree:
            stripped_dir = dirname.replace(options.basedir, '')
            if not options.case_sensitive:
                stripped_dir = stripped_dir.lower()

            dir_components = stripped_dir.split(os.path.sep)
            for ignore in self.ignore_dirs:
                if not options.case_sensitive:
                    ignore = ignore.lower()
                if ignore in dir_components:
                    break
            else:
                rv[dirname] = tree[dirname]
        return rv


class PruneEmpty(BaseFilter):
    def __init__(self):
        super().__init__('UNUSED')

    def apply(self, tree, _):
        rv = {}
        for dirname in tree:
            if tree[dirname]:
                rv[dirname] = tree[dirname]
        return rv

class StandardMatch(BaseFilter):
    def apply(self, tree, options):
        if not options.case_sensitive:
            this_filter = self.filter_string.lower()
        else:
            this_filter = self.filter_string
        rv = {}
        for dirname in tree:
            if tree[dirname]:
                rv[dirname] = list()
                for filename in tree[dirname]:
                    if '/' in this_filter:
                        # print('all up in dis filter')
                        match_name = os.path.join(dirname, filename)
                        if options.basedir:
                            match_name = match_name.replace(options.basedir, '')
                        # print(match_name)
                    else:
                        match_name = filename
                    if not options.case_sensitive:
                        match_name = match_name.lower()
                    if this_filter in match_name:
                        rv[dirname].append(filename)
            else:
                rv[dirname] = None
        return rv


class FuzzyMatch(BaseFilter):
    def apply(self, tree, options):
        if not options.case_sensitive:
            this_filter = self.filter_string.lower()
        else:
            this_filter = self.filter_string
        rv = {}
        for dirname in tree:
            if tree[dirname]:
                rv[dirname] = list()
                for filename in tree[dirname]:
                    if options.fuzzy_include_dir:
                        match_name = os.path.join(dirname, filename)
                        if options.fuzzy_strip_basedir and options.basedir:
                            match_name = match_name.replace(options.basedir, '')
                    else:
                        match_name = filename
                    if not options.case_sensitive:
                        match_name = match_name.lower()
                    index = 0
                    for letter in this_filter:
                        index = match_name.find(letter, index)
                        if index == -1:
                            break
                    else:
                        rv[dirname].append(filename)
            else:
                rv[dirname] = None
        return rv


class StartsWith(BaseFilter):
    def apply(self, tree, options):
        if not options.case_sensitive:
            this_filter = self.filter_string.lower()
        else:
            this_filter = self.filter_string
        rv = {}
        for dirname in tree:
            if tree[dirname]:
                rv[dirname] = list()
                for filename in tree[dirname]:
                    if not options.case_sensitive:
                        match_name = filename.lower()
                    else:
                        match_name = filename
                    if match_name.startswith(this_filter):
                        rv[dirname].append(filename)
            else:
                rv[dirname] = None
        return rv


class EndsWith(BaseFilter):
    def apply(self, tree, options):
        if not options.case_sensitive:
            this_filter = self.filter_string.lower()
        else:
            this_filter = self.filter_string
        rv = {}
        for dirname in tree:
            if tree[dirname]:
                rv[dirname] = list()
                for filename in tree[dirname]:
                    if not options.case_sensitive:
                        match_name = filename.lower()
                    else:
                        match_name = filename
                    if match_name.endswith(this_filter):
                        rv[dirname].append(filename)
            else:
                rv[dirname] = None
        return rv



import json
import os

import click

from ndl import utils, filters

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


def json_out(tree, options):
    click.echo(json.dumps(tree))


def flat_out(tree, options):
    for x in sorted(tree.keys()):
        for f in tree[x]:
            f_out = os.path.join(x, f)
            if not options.show_full_path:
                f_out = f_out.replace(options.basedir + os.path.sep, '')
            click.echo(f_out)

OUTPUT_FUNCS = {
    'json': json_out,
    'flat': flat_out,
}

def err_out(msg):
    click.echo(msg, err=True)


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('path', default='.', type=click.Path(exists=True))
@click.option('-s', '--startswith', default=None)
@click.option('-e', '--endswith', default=None)
@click.option('-f', '--fuzzy', default=None, multiple=True)
@click.option('-i', '--ignore', default=None, multiple=True)
@click.option('--show-full-path/--hide-full-path', default=True)
@click.option('--fuzzy-include-dir', is_flag=True)
@click.option('-o', '--out-type', default='flat', type=click.Choice(['json', 'flat']))
def cli(
        path,
        startswith,
        endswith,
        fuzzy,
        ignore,
        out_type,
        show_full_path,
        fuzzy_include_dir):

    filter_options = utils.FilterOptions(
        fuzzy_include_dir=fuzzy_include_dir,
        show_full_path=show_full_path,
    )
    path = os.path.realpath(path)
    base_tree = utils.build_tree(path, filter_options, err_out=err_out)
    user_filters = []
    if ignore is not None:
        user_filters.extend((filters.IgnoreDir(x) for x in ignore))
    if endswith is not None:
        user_filters.append(filters.EndsWith(endswith))
        # user_filters.extend([filters.EndsWith(x) for x in  endswith])
    if startswith is not None:
        user_filters.append(filters.StartsWith(startswith))
        # user_filters.extend([filters.StartsWith(x) for x in startswith])
    if fuzzy is not None:
        user_filters.extend([filters.FuzzyMatch(x) for x in fuzzy])

    filtered_tree = utils.apply_filters(base_tree, user_filters, filter_options)
    filter_options.store_filters(user_filters)
    OUTPUT_FUNCS[out_type](filtered_tree, filter_options)
    filter_options.save()



import os
from . import filters as ndl_filters


class FilterOptions(object):
    def __init__(self,
                 case_sensitive=False,  # Applies to all matching filters
                 fuzzy_include_dir=False,  # include dirname in fuzzy filters
                 fuzzy_strip_basedir=True,  # ... but strip common ancestor dir
                 prune_empty_dirs=True,  # remove empty dirs from final tree
                 hide_dotfiles=True,  # don't show hidden files
                 show_full_path=False,  # output files with absolute paths
                 ):
        self.case_sensitive = case_sensitive
        self.fuzzy_include_dir = fuzzy_include_dir
        self.fuzzy_strip_basedir = fuzzy_strip_basedir
        self.prune_empty_dirs = prune_empty_dirs
        self.hide_dotfiles = hide_dotfiles
        self.show_full_path = show_full_path


        self.basedir = None
        self.filters = None

    def set_basedir(self, basedir):
        self.basedir = basedir

    def store_filters(self, filters):
        self.filters = filters

    def save(self):
        pass


def build_tree(cwd=None, options=None, err_out=None):
    if cwd is None:
        cwd = os.getcwd()

    if options:
        options.set_basedir(cwd)
        hide_dots = options.hide_dotfiles
    else:
        hide_dots = False

    def walk(d):
        files = []
        dirs = []
        try:
            for f in os.listdir(d):
                if hide_dots and f.startswith('.'):
                    continue
                if os.path.isdir(os.path.join(d, f)):
                    dirs.append(os.path.join(d, f))
                else:
                    files.append(f)
        except PermissionError:
            if err_out:
                err_out('Access Denied: {}'.format(d))
        rv = {d: files}
        for sub_dir in dirs:
            rv.update(walk(sub_dir))
        return rv
    return walk(cwd)


def apply_filters(tree, filters, options):
    if not filters:
        return tree

    filtered = tree.copy()
    # filters = sorted(filters)
    if options.prune_empty_dirs:
        filters.append(ndl_filters.PruneEmpty())
    for f in filters:
        filtered = f.apply(filtered, options)

    return filtered


def print_tree(tree, options):
    keys = sorted(tree.keys())
    print('Directory: {}\n'.format(options.basedir))
    for key in keys:
        stripped_key = key.replace(options.basedir, '')
        stripped_key = '/{}/'.format(stripped_key)
        stripped_key = stripped_key.replace('//', '/')
        depth = stripped_key.count('/') - 1
        stripped_key = stripped_key.split('/')[-2].strip()
        print('{}{}/'.format(('  ' * depth), stripped_key))
        for filename in tree[key]:
            print('{}{}'.format(('  ' * (depth+1)), filename))





from setuptools import setup

__VERSION__ = '0.1'

setup(
    name='ndl',
    packages=['ndl'],
    py_modules=['ndl'],
    description='Easy file searching from the command-line',
    author='Sully Sullenberger',
    author_email='sully@sadburger.com',
    url='https://github.com/msull/needle',
    download_url='https://github.com/msull/needle/tarball/%s' % __VERSION__,
    keywords=['search', 'files', 'terminal'],
    classifiers=[],
    version=__VERSION__,
    install_requires=[
        'Click',
    ],
    entry_points="""
        [console_scripts]
        ndl=ndl.cli:cli
    """,
)
